# Copyright (C) 2012 Valentin Rothberg <valentinrothberg@googlemail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

class Config(dict):
    def __init__(self, path = None):
        dict.__init__(self)
        if path != None:
            self.readConfigFile(path)

    def readConfigFile(self, path):
        """
        Parse the file in path and add all features contained to the config.
        """
        with open(path, 'r') as stream:
            for line in stream.readlines():
                if line.startswith("CONFIG_"):
                    line = line.strip("\n")
                    split = line.split("=")
                    key = split[0].strip()
                    value = split[1].strip()
                    self.update({key:value})

    def valueOf(self, key):
        """
        Return value of key. If key is not present in the config, None is
        returned.
        """
        if key in self.keys():
            return self[key]
        else:
            return None

    def addFeature(self, key, value):
        """
        Add feature as key and value to the config.
        """
        self.update({key:value})

    def contains(self, key):
        """
        Return true if config contains key.
        """
        return key in self.keys()

    def conflict(self, other):
        """
        Return true if both Configs conflict. Configs conflict if both contain
        the same feature with different values.
        """
        for key in self.keys():
            if other.contains(key):
                if self.valueOf(key) != other.valueOf(key):
                    return True

    def getConflicts(self, other):
        """
        Return a list of conflicting features in both configs.
        """
        res = []
        for key in self.keys():
            if other.contains(key):
                if self.valueOf(key) != other.valueOf(key):
                    res.append(key)
        return res



"""
Attributes:

   default_architecture   -- the default architecture to use
   prefer_32bit  -- If running on an amd64 64bit host, prefer configuring and compiling for 32bit.
   kernelversion -- normally determined automatically, but can be overridden here
"""

default_architecture = "x86"

prefer_32bit = True

kernelversion = None




#
#   rsf2model - extracts presence implications from kconfig dumps
#
# Copyright (C) 2011 Christian Dietrich <christian.dietrich@informatik.uni-erlangen.de>
# Copyright (C) 2012 Manuel Zerpies <manuel.f.zerpies@ww.stud.uni-erlangen.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from vamos.rsf2model import tools
from vamos.rsf2model.BoolParser import BoolParser
from vamos.rsf2model.helper import tree_change, BoolRewriterException

class BoolRewriter(tools.UnicodeMixin):
    ELEMENT = "in"

    def __init__(self, rsf, expr, eval_to_module = True):
        tools.UnicodeMixin.__init__(self)
        self.expr = BoolParser(expr).to_bool()
        if type(self.expr) == str or self.expr[0] == BoolParser.NOT:
            self.expr = [BoolParser.AND, self.expr]
        self.rsf = rsf
        self.eval_to_module = eval_to_module

    def rewrite_not(self):
        self.expr = tree_change(self.__rewrite_not, self.expr)
        return self.expr

    def __rewrite_not(self, tree):
        if tree[0] == BoolParser.NOT and type(tree[1]) == list:
            tree = tree[1]
            if tree[0] == BoolParser.AND:
                tree = [BoolParser.OR] + map(lambda x: [BoolParser.NOT, x], tree[1:])
                return tree_change(self.__rewrite_not, tree)
            elif tree[0] == BoolParser.OR:
                tree = [BoolParser.AND] + map(lambda x: [BoolParser.NOT, x], tree[1:])
                return tree_change(self.__rewrite_not, tree)
            elif tree[0] == BoolParser.NOT:
                return tree_change(self.__rewrite_not, tree[1])
            elif tree[0] == BoolParser.EQUAL:
                tree[0] = BoolParser.NEQUAL
                return tree
            elif tree[0] == BoolParser.NEQUAL:
                tree[0] = BoolParser.EQUAL
                return tree

    def rewrite_tristate(self):
        self.expr = tree_change(self.__rewrite_tristate, self.expr)
        return self.expr

    def __rewrite_tristate(self, tree):
        def tristate_not(symbol):
            if symbol in self.rsf.options() and self.rsf.options()[symbol].tristate():
                if self.eval_to_module:
                    return [BoolParser.NEQUAL, symbol, "y"]
                else:
                    return [BoolParser.EQUAL,  symbol, "n"]
            return [BoolParser.NOT, symbol]
        def tristate(symbol):
            if symbol in self.rsf.options() and self.rsf.options()[symbol].tristate():
                if self.eval_to_module:
                    return [BoolParser.NEQUAL, symbol, "n"]
                else:
                    return [BoolParser.EQUAL,  symbol, "y"]
            return symbol

        if tree[0] in [BoolParser.AND, BoolParser.OR]:
            for i in range(1, len(tree)):
                if type(tree[i]) == list:
                    if tree[i][0] == BoolParser.NOT:
                        tree[i] = tristate_not(tree[i][1])
                    else:
                        tree[i] = tree_change(self.__rewrite_tristate, tree[i])
                else:
                    tree[i] = tristate(tree[i])
            return tree

    def rewrite_choice(self):
        """Removes all CHOICE_ items"""
        def __recr(tree):
            tree = filter(lambda x: not(type(x) == str and x.startswith("CHOICE_")), tree)
            if len(tree) == 1:
                return []
            return tree

        self.expr = tree_change(__recr, self.expr)
        return self.expr

    def rewrite_symbol(self):
        self.expr = tree_change(self.__rewrite_symbol, self.expr)
        return self.expr

    def __rewrite_symbol(self, tree):
        def to_symbol(tree):
            if type(tree) in [list, tuple]:
                return tree_change(self.__rewrite_symbol, tree)
            if tree == "m":
                if self.eval_to_module:
                    # m is true, if the expression can evaluate to module
                    return tools.new_free_item()
                else:
                    #otherwise it is false, because expr = y is needed
                    a = tools.new_free_item()
                    return [BoolParser.AND, a,
                            [BoolParser.NOT, a]]
            return self.rsf.symbol(tree)

        if tree[0] in [BoolParser.NOT, BoolParser.AND, BoolParser.OR]:
            return [tree[0]] + map(to_symbol, tree[1:])
        elif tree[0] == BoolParser.EQUAL:
            return self.__rewrite_symbol_equal(tree)
        elif tree[0] == BoolParser.NEQUAL:
            return self.__rewrite_symbol_nequal(tree)

    def __rewrite_symbol_equal(self,tree):
        left = tree[1]
        right = tree[2]
        left_y = self.rsf.symbol(left)
        left_m = self.rsf.symbol_module(left)
        right_y = self.rsf.symbol(right)
        right_m = self.rsf.symbol_module(right)

        if left.lower() in ["y", "n", "m"]:
            right, left = left, right
        if left.lower() in ["y", "n", "m"]:
            raise BoolRewriterException("compare literal with literal")
        if right == "y":
            return left_y
        elif right == "m":
            return left_m
        elif right == "n":
            return [BoolParser.AND,
                    [BoolParser.NOT, left_m],
                    [BoolParser.NOT, left_y]]
        else:
            # Symbol == Symbol
            return [BoolParser.OR,
                    [BoolParser.AND, left_y, right_y], # Either both y
                    [BoolParser.AND, left_m, right_m], # Or both
                    [BoolParser.AND, # Or everything disabled
                     [BoolParser.NOT, left_y], [BoolParser.NOT, right_y],
                     [BoolParser.NOT, left_m], [BoolParser.NOT, right_m]]]


    def __rewrite_symbol_nequal(self,tree):
        left = tree[1]
        right = tree[2]
        left_y = self.rsf.symbol(left)
        left_m = self.rsf.symbol_module(left)
        right_y = self.rsf.symbol(right)
        right_m = self.rsf.symbol_module(right)

        if left.lower() in ["y", "n", "m"]:
            right, left = left, right
        if left.lower() in ["y", "n", "m"]:
            raise BoolRewriterException("compare literal with literal")
        if right == "y":
            return [BoolParser.NOT, left_y]
        elif right == "m":
            return [BoolParser.NOT, left_m]
        elif right == "n":
            return [BoolParser.OR, left_m, left_y]
        else:
            return [BoolParser.OR,
                    [BoolParser.AND, left_y, [BoolParser.NOT, right_y]],
                    [BoolParser.AND, left_m, [BoolParser.NOT, right_m]],
                    [BoolParser.AND, [BoolParser.NOT, left_y], right_y],
                    [BoolParser.AND, [BoolParser.NOT, left_m], right_m]]


    def dump(self):
        def __concat(tree):
            if not type(tree) in [list, tuple]:
                return tree
            if tree[0] == BoolParser.NOT:
                return "!" + tree[1]
            elements = []
            for e in tree[1:]:
                elements.append(__concat(e))
            cat = " && "
            if tree[0] == BoolParser.OR:
                cat = " || "
            if len(elements) == 1:
                return cat.join(elements)
            return "(" + cat.join(elements) + ")"
        if self.expr == []:
            return ""
        return __concat(self.expr)

    def rewrite(self):
        self.rewrite_not()
        self.rewrite_choice()
        self.rewrite_tristate()
        self.rewrite_symbol()
        return self

    def __unicode__(self):
        return self.dump()



#
#   vamos - common auxiliary functionality
#
# Copyright (C) 2011 Christian Dietrich <christian.dietrich@informatik.uni-erlangen.de>
# Copyright (C) 2011 Reinhard Tartler <tartler@informatik.uni-erlangen.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import copy

class Selection:
    """A selection is a selection of symbols (derived from kconfig
    features). """

    def __init__(self, other_selection = None):
        """A selection can be empty, or can be copied from another
        selection. So this is somewhat like a copy constructor in c++.

        other_selection == None: empty selection
        other_selection == [(),()]: selection from tuple
        other_selection == Selection(..): copy selection from other selection"""

        # expr == (SYMBOL || SYMBOL) && (SYMBOL || SYMBOL)
        # inner sets: alternatives
        # outer list: conjugations

        self.expr = [set([])]

        # All mentioned symbols in this selection
        self.symbols = set([])

        if other_selection:
            # Copy the other expression
            if type(other_selection) == list:
                self.expr = copy.deepcopy(other_selection)
            else:
                self.expr = copy.deepcopy(other_selection.expr)

        for or_expr in self.expr:
            for symbol in or_expr:
                self.symbols.add(symbol)

    def push_down(self):
        """Add a new level of or_expressions to the conjugation"""
        self.expr.append(set([]))


    def add_alternative(self, alternative):
        """Add an alternative to the current "top of stack" or expression"""
        self.expr[-1].add(alternative)
        self.symbols.add(alternative)


    def features(self):
        """From a list of symbols determine the features and their selection:

        CONFIG_BAR -> CONFIG_BAR=y
        CONFIG_FOO_MODULE -> CONFIG_FOO=m"""
        features = {}
        for or_expr in self.expr:
            if len(or_expr) == 0:
                continue
            symbol = list(or_expr)[0]
            value = "y"
            if symbol.endswith("_MODULE"):
                symbol = symbol[:-len("_MODULE")]
                value = "m"
            features[symbol] = value
        return features


    def feature_in_selection(self, feature):
        """ Tests if feature ({,_MODULE}) is already in this selection"""
        if feature.endswith("_MODULE"):
            feature = feature[:-len("_MODULE")]
        return feature in self.symbols or \
               feature + "_MODULE" in self.symbols


    def __str__(self):
        """Format the selection to a (normalized) propositional formula
        (return-type: string)"""
        or_exprs = []
        for or_expr in self.expr:
            if len(or_expr) == 0:
                continue
            elif len(or_expr) == 1:
                or_exprs.append(list(or_expr)[0])
            else:
                or_exprs.append("(" + " || ".join(sorted(list(or_expr))) + ")")

        if len(or_exprs) > 0:
            return " && ".join(sorted(or_exprs))
        else:
            return ""

    __repr__ = __str__


    def better_than(self, other_selection):
        """One Selection is better than another if it is a subset of
        the other (and if it is shorter)"""
        if len(self.expr) < len(other_selection.expr):
            # When our expression is shorter we can test if our
            # expression is a subset of the other expression

            # This means all our expressions are within the other
            # expression
            for or_expr in self.expr:
                within = False
                for or_expr_1 in other_selection.expr:
                    if or_expr_1 == or_expr:
                        within = True
                if within == False:
                    return False
            return True
        else:
            return False


    def merge(self, other):
        """Expressions can be merged into one selection if they differ
        only in or_expression"""
        if len(self.expr) != len(other.expr):
            return None

        new_expr = []
        differences = 0
        different_or_expr = set()

        for or_expr in self.expr:
            within = False
            for or_expr_1 in other.expr:
                if or_expr_1 == or_expr:
                    new_expr.append(or_expr)
                    within = True
            if not within:
                differences += 1
                different_or_expr = different_or_expr.union(or_expr)
            if differences > 1:
                return None

        for or_expr in other.expr:
            within = False
            for or_expr_1 in self.expr:
                if or_expr_1 == or_expr:
                    within = True
            if not within:
                different_or_expr = different_or_expr.union(or_expr)

        new_expr.append(different_or_expr)

        return Selection(new_expr)
